# Copyright 2021 The SQLNet Company GmbH

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from datetime import datetime
from os import makedirs, listdir
from os.path import isdir
from pathlib import Path
from platform import system
import socket
from subprocess import Popen
from time import sleep
from typing import List, NamedTuple, Optional, Union

from getml.communication import tcp_port
from getml.version import __version__


class _Options(NamedTuple):
    allow_push_notifications: bool
    allow_remote_ips: bool
    home_directory: Optional[str]
    http_port: Optional[int]
    in_memory: bool
    install: bool
    launch_browser: bool
    log: bool
    project_directory: Optional[str]
    proxy_url: Optional[str]
    token: Optional[str]

    def to_cmd(self, binary_path: Path) -> List[str]:
        """
        Generates the command to be passed to Popen.
        """
        return [str(binary_path)] + [
            "--" + key.replace("_", "-") + "=" + _handle_value(value)
            for (key, value) in self._asdict().items()  # pylint: disable=E1101
            if value is not None
        ]


def _find_binary(home_directory: Path) -> Path:
    getml_directory = home_directory / ".getML"
    version_directories = [
        d
        for d in listdir(getml_directory)
        if isdir(getml_directory / d) and "getml-" + __version__ in str(d)
    ]
    if not version_directories:
        raise OSError(
            "getml-"
            + __version__
            + " is not installed in "
            + str(getml_directory)
            + "."
        )
    if len(version_directories) > 1:
        raise OSError(
            "More than one installation found for getml-"
            + __version__
            + " in "
            + str(getml_directory)
            + "."
        )
    if system() == "Linux":
        return getml_directory / version_directories[0] / "getML"
    if system() == "Windows":
        return getml_directory / version_directories[0] / "getML.exe"
    return getml_directory / version_directories[0] / "getml-cli"


def _handle_bool(value: bool) -> str:
    if value:
        return "true"
    return "false"


def _handle_value(value: Union[str, int, bool]) -> str:
    if isinstance(value, bool):
        return _handle_bool(value)
    return str(value)


def _make_log_path(home_directory: Path) -> Path:
    getml_directory = home_directory / ".getML"
    makedirs(getml_directory / "logs", exist_ok=True)
    return getml_directory / "logs" / datetime.now().strftime("%Y%m%d%H%M%S.log")


def _is_monitor_alive():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect(("localhost", tcp_port))
    except ConnectionRefusedError:
        return False
    return True


def launch(
    allow_push_notifications=True,
    allow_remote_ips=False,
    home_directory: Optional[str] = None,
    http_port: Optional[int] = None,
    in_memory=True,
    install=False,
    launch_browser=True,
    log=False,
    project_directory: Optional[str] = None,
    proxy_url: Optional[str] = None,
    token: Optional[str] = None,
):
    """
    Launches the getML engine.

    Args:
      allow_push_notifications (bool):
            Whether you want to allow the getML monitor to send push notifications to your desktop.

      allow_remote_ips (bool):
            Whether you want to allow remote IPs to access the http-port.

      home_directory (str, optional):
            The directory which should be treated as the home directory by getML.
            getML will create a hidden folder named '.getML' in said directory.
            All binaries will be installed there.

      http_port (int, optional):
            The local port of the getML monitor.
            This port can only be accessed from your local computer,
            unless you set allow_remote_ips=True.

      in_memory (bool):
            Whether you want the engine to process everything in memory.

      install (bool)
            Reinstalls getML, even if it is already installed.

      launch_browser (bool):
            Whether you want to automatically launch your browser.

      log (bool):
            Whether you want the engine log to appear in the logfile (more detailed logging).
            The engine log also appears in the 'Log' page of the monitor.

      project_directory (str, optional):
            The directory in which to store all of your projects.

      proxy_url (str, optional):
            The URL of any proxy server that that redirects to the getML monitor.

      tcp_port (int, optional):
            Local TCP port which serves as the communication point for the engine.
            This port can only be accessed from your local computer.

      token (str, optional):
            The token used for authentication.
            Authentication is required when remote IPs are allowed to access the monitor.
            If authentication is required and no token is passed,
            a random hexcode will be generated as the token."""
    if _is_monitor_alive():
        print("getML engine is already running.")
        return
    home_path = Path(home_directory) if home_directory is not None else Path.home()
    binary_path = _find_binary(home_path)
    log_path = _make_log_path(home_path)
    logfile = open(str(log_path), "w", encoding="utf-8")
    cmd = _Options(
        allow_push_notifications=allow_push_notifications,
        allow_remote_ips=allow_remote_ips,
        home_directory=home_directory,
        http_port=http_port,
        in_memory=in_memory,
        install=install,
        launch_browser=launch_browser,
        log=log,
        project_directory=project_directory,
        proxy_url=proxy_url,
        token=token,
    ).to_cmd(binary_path)
    Popen(cmd, shell=False, stdout=logfile, stdin=logfile, stderr=logfile)
    while not _is_monitor_alive():
        sleep(0.1)
    print(
        "Launched the getML engine. The log output will be stored in "
        + str(log_path)
        + "."
    )
